@@ -1,6 +1,6 @@
 /* External client API, using UNIX domain socket.
  *
- * Copyright (c) 2015-2021  Joachim Wiberg <troglobit@gmail.com>
+ * Copyright (c) 2015-2024  Joachim Wiberg <troglobit@gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -21,13 +21,13 @@
  * THE SOFTWARE.
  */
 
-#include <err.h>
 #include <errno.h>
 #include <poll.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 
 #include "client.h"
+#include "log.h"
 
 static int sd = -1;
 
@@ -39,11 +39,17 @@ int client_connect(void)
 	};
 
 	sd = socket(AF_UNIX, SOCK_SEQPACKET, 0);
-	if (-1 == sd)
-		err(1, "Failed creating UNIX domain socket");
+	if (-1 == sd) {
+		warn("Failed creating UNIX domain socket");
+		return -1;
+	}
 
-	if (connect(sd, (struct sockaddr*)&sun, sizeof(sun)) == -1)
-		err(1, "Failed connecting to finit");
+	if (connect(sd, (struct sockaddr*)&sun, sizeof(sun)) == -1) {
+		if (errno != ENOENT)
+			warnx("Failed connecting to finit");
+		close(sd);
+		sd = -1;
+	}
 
 	return sd;
 }
@@ -63,64 +69,97 @@ int client_disconnect(void)
 	return rc;
 }
 
-int client_send(struct init_request *rq, ssize_t len)
+/* Only valid after client_connect() */
+int client_socket(void)
+{
+	return sd;
+}
+
+int client_request(struct init_request *rq, ssize_t len)
 {
 	struct pollfd pfd = { 0 };
-	int sd, result = 255;
 	int rc;
 
-	sd = client_connect();
-	if (-1 == sd)
+	if (client_connect() == -1)
 		return -1;
 
 	pfd.fd     = sd;
 	pfd.events = POLLOUT;
 	if (poll(&pfd, 1, 2000) <= 0) {
 		warn("Timed out waiting for Finit, errno %d", errno);
-		goto exit;
+		return -1;
 	}
 
 	if (write(sd, rq, len) != len) {
 		warn("Failed communicating with Finit, errno %d", errno);
-		goto exit;
+		return -1;
+	}
+
+	/*
+	 * Exception, only command that don't need to wait for a
+	 * response from Finit.  Registers the client socket as a
+	 * readiness notification socket with Finit
+	 */
+	if (rq->cmd == INIT_CMD_NOTIFY_SOCKET) {
+		rq->cmd = INIT_CMD_ACK;
+		return 0;
 	}
 
 	pfd.fd = sd;
-	pfd.events = POLLIN;
+	pfd.events = POLLIN | POLLERR | POLLHUP;
 	if ((rc = poll(&pfd, 1, 2000)) <= 0) {
-		if (rc)
+		if (rc) {
+			if (errno == EINTR) /* shutdown/reboot */
+				return -1;
 			warn("poll(), errno %d", errno);
-		else
+		} else
 			warnx("Timed out waiting for reply from Finit.");
-		goto exit;
+		return -1;
 	}
 
 	if (read(sd, rq, len) != len) {
 		warn("Failed reading reply from Finit, errno %d", errno);
-		goto exit;
+		return -1;
 	}
 
 	if (rq->cmd == INIT_CMD_NACK)
-		result = 1;
-	else
-		result = 0;
-exit:
+		return -1;
+
+	return 0;
+}
+
+int client_send(struct init_request *rq, ssize_t len)
+{
+	int rc = 0;
+
+	if (client_request(rq, len))
+		rc = 255;
+
 	client_disconnect();
 
-	return result;
+	return rc;
+}
+
+int client_command(int cmd)
+{
+	struct init_request rq = {
+		.magic = INIT_MAGIC,
+		.runlevel = getpid(),
+		.cmd = cmd,
+	};
+
+	return client_request(&rq, sizeof(rq));
 }
 
 svc_t *client_svc_iterator(int first)
 {
-	int sd = -1;
 	struct init_request rq = {
 		.magic = INIT_MAGIC,
 		.cmd   = INIT_CMD_SVC_ITER,
 	};
 	static svc_t svc;
 
-	sd = client_connect();
-	if (sd == -1)
+	if (client_connect() == -1)
 		return NULL;
 
 	if (first)
@@ -139,24 +178,22 @@ svc_t *client_svc_iterator(int first)
 
 	return &svc;
 error:
-	perror("Failed communicating with finit");
+	warn("Failed communicating with finit, error %d", errno);
 	client_disconnect();
 	sd = -1;
 
 	return NULL;
 }
 
-svc_t *do_cmd(int cmd, const char *arg)
+static svc_t *do_find(int cmd, const char *arg)
 {
 	struct init_request rq = {
 		.magic = INIT_MAGIC,
 		.cmd   = cmd,
 	};
 	static svc_t svc;
-	int sd = -1;
 
-	sd = client_connect();
-	if (sd == -1)
+	if (client_connect() == -1)
 		return NULL;
 
 	strlcpy(rq.data, arg, sizeof(rq.data));
@@ -172,20 +209,19 @@ svc_t *do_cmd(int cmd, const char *arg)
 	return &svc;
 error:
 	client_disconnect();
-	perror("Failed communicating with finit");
+	warn("Failed communicating with finit, error %d", errno);
 
 	return NULL;
 }
 
 svc_t *client_svc_find(const char *arg)
 {
-	return do_cmd(INIT_CMD_SVC_FIND, arg);
+	return do_find(INIT_CMD_SVC_FIND, arg);
 }
 
-
 svc_t *client_svc_find_by_cond(const char *arg)
 {
-	return do_cmd(INIT_CMD_SVC_FIND_BYC, arg);
+	return do_find(INIT_CMD_SVC_FIND_BYC, arg);
 }
 
 /**
