@@ -1,6 +1,6 @@
 /* New client tool, replaces old /dev/initctl API and telinit tool
  *
- * Copyright (c) 2015-2021  Joachim Wiberg <troglobit@gmail.com>
+ * Copyright (c) 2015-2024  Joachim Wiberg <troglobit@gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -23,13 +23,13 @@
 
 #include "config.h"
 
-#include <err.h>
 #include <ftw.h>
 #include <ctype.h>
 #include <getopt.h>
 #include <paths.h>
 #include <signal.h>
 #include <stdio.h>
+#include <syslog.h>
 #include <time.h>
 #include <utmp.h>
 #include <arpa/inet.h>
@@ -39,6 +39,7 @@
 # include <lite/lite.h>
 #endif
 
+#include "initctl.h"
 #include "client.h"
 #include "cond.h"
 #include "serv.h"
@@ -52,17 +53,24 @@ struct cmd {
 	struct cmd  *ctx;
 	int        (*cb)(char *arg);
 	int         *cond;
+	int        (*cb_multiarg)(int argc, char **argv);
 };
 
+char *finit_conf = NULL;
+char *finit_rcsd = NULL;
+
 int icreate  = 0;
 int iforce   = 0;
 int ionce    = 0;
 int debug    = 0;
 int heading  = 1;
+int json     = 0;
+int noerr    = 0;
 int verbose  = 0;
 int plain    = 0;
 int quiet    = 0;
 int runlevel = 0;
+int timeout  = 0;
 int cgrp     = 0;
 int utmp     = 0;
 
@@ -95,8 +103,8 @@ static void col_widths(void)
 	}
 
 	/* adjust for min col width */
-	if (iw < 5)
-		iw = 5;
+	if (iw < 6)
+		iw = 6;
 	if (pw < 3)
 		pw = 3;
 }
@@ -153,17 +161,42 @@ static int toggle_debug(char *arg)
 	return client_send(&rq, sizeof(rq));
 }
 
-static int do_log(char *svc)
+static int do_log(svc_t *svc, char *tail)
 {
-	char *logfile = "/var/log/syslog";
+	const char *logfile = "/var/log/syslog";
+	pid_t pid;
+	char *nm;
 
-	if (!svc || !svc[0])
-		svc = "finit";
+	if (svc) {
+		nm = svc_ident(svc, NULL, 0);
+		pid = svc->pid;
+		if (!pid)
+			return 0; /* not running */
+	} else {
+		nm  = "finit";
+		pid = 1;
+	}
 
-	if (!fexist(logfile))
+	if (!fexist(logfile)) {
 		logfile = "/var/log/messages";
+		if (!fexist(logfile))
+			return 0; /* bail out, maybe in container */
+	}
 
-	return systemf("cat %s | grep %s | tail -10", logfile, svc);
+	return systemf("cat %s | grep '\\[%d\\]\\|%s' %s", logfile, pid, nm, tail);
+}
+
+static int show_log(char *arg)
+{
+	svc_t *svc = NULL;
+
+	if (arg) {
+		svc = client_svc_find(arg);
+		if (!svc)
+			ERRX(noerr ? 0 : 69, "no such task or service(s): %s", arg);
+	}
+
+	return do_log(svc, "");
 }
 
 static int do_runlevel(char *arg)
@@ -176,19 +209,26 @@ static int do_runlevel(char *arg)
 	if (!arg) {
 		int prevlevel = 0;
 		int currlevel;
-		char prev;
+		char prev, curr;
 
 		currlevel = runlevel_get(&prevlevel);
-		if (255 == currlevel) {
+		switch (currlevel) {
+		case 255:
 			printf("unknown\n");
 			return 0;
+		case INIT_LEVEL:
+			curr = 'S';
+			break;
+		default:
+			curr = currlevel + '0';
+			break;
 		}
 
 		prev = prevlevel + '0';
 		if (prev <= '0' || prev > '9')
 			prev = 'N';
 
-		printf("%c %d\n", prev , currlevel);
+		printf("%c %c\n", prev , curr);
 		return 0;
 	}
 
@@ -225,20 +265,11 @@ static int do_startstop(int cmd, char *arg)
 	};
 
 	if (!arg || !arg[0])
-		errx(1, "missing argument to %s", (cmd == INIT_CMD_START_SVC
-						   ? "start"
-						   : (cmd == INIT_CMD_STOP_SVC
-						      ? "stop"
-						      : "restart")));
+		ERRX(2, "missing command argument");
 
 	strlcpy(rq.data, arg, sizeof(rq.data));
-	if (client_send(&rq, sizeof(rq))) {
-		fprintf(stderr, "No such task or service(s): %s\n\n", arg);
-		fprintf(stderr, "Usage: initctl %s <NAME>[:ID]\n",
-			cmd == INIT_CMD_START_SVC ? "start" :
-			(cmd == INIT_CMD_STOP_SVC ? "stop"  : "restart"));
-		return 1;
-	}
+	if (client_send(&rq, sizeof(rq)))
+		ERRX(noerr ? 0 : 69, "no such task or service(s): %s", arg);
 
 	return do_svc(cmd, arg);
 }
@@ -256,25 +287,57 @@ static int do_reload (char *arg)
 
 static int do_restart(char *arg)
 {
-	size_t retries = 3;
-	svc_t *svc;
+	if (do_startstop(INIT_CMD_RESTART_SVC, arg))
+		ERRX(noerr ? 0 : 7, "failed restarting %s", arg);
 
-	if (do_startstop(INIT_CMD_STOP_SVC, arg))
-		return 1;
-
-	while (retries-- > 0 && (svc = client_svc_find(arg))) {
-		if (!svc_is_running(svc))
-			break;
-
-		sleep(1);
-	}
-
-	if (retries == 0)
-		errx(1, "Failed stopping %s (restart)", arg);
-
-	return do_startstop(INIT_CMD_RESTART_SVC, arg);
+	return 0;
 }
 
+/**
+ * do_signal - Ask finit to send a signal to a service.
+ * @argv: must point to an array of strings, containing a service
+ *        and signal name, in that order.
+ * @argc: must be 2.
+ *
+ * A signal can be a complete signal name such as "SIGHUP", or
+ * it can be the shortest unique name, such as "HUP" (no SIG prefix).
+ * It can also be a raw signal number, such as "9" (SIGKILL).
+ */
+int do_signal(int argc, char *argv[])
+{
+	struct init_request rq = {
+		.magic = INIT_MAGIC,
+		.cmd   = INIT_CMD_SVC_QUERY
+	};
+	int signo;
+
+	if (argc != 2)
+		ERRX(2, "invalid number of arguments to signal");
+
+	strlcpy(rq.data, argv[0], sizeof(rq.data));
+	if (client_send(&rq, sizeof(rq)))
+		ERRX(noerr ? 0 : 69, "no such task or service(s): %s", argv[0]);
+
+	signo = str2sig(argv[1]);
+	if (signo == -1) {
+		const char *errstr = NULL;
+
+		signo = (int)strtonum(argv[1], 1, 31, &errstr);
+		if (errstr)
+			ERRX(65, "%s signal: %s", errstr, argv[1]);
+	}
+
+	/* Reuse runlevel for signal number. */
+	rq.magic    = INIT_MAGIC;
+	rq.cmd      = INIT_CMD_SIGNAL;
+	rq.runlevel = signo;
+	strlcpy(rq.data, argv[0], sizeof(rq.data));
+
+	return client_send(&rq, sizeof(rq));
+}
+
+int dump_once;
+char *dump_filter;
 static int dump_one_cond(const char *fpath, const struct stat *sb, int tflag, struct FTW *ftwbuf)
 {
 	const char *cond, *asserted;
@@ -289,10 +352,14 @@ static int dump_one_cond(const char *fpath, const struct stat *sb, int tflag, st
 
 	asserted = condstr(cond_get_path(fpath));
 	cond = &fpath[strlen(_PATH_COND)];
+
+	if (dump_filter && dump_filter[0] && strncmp(cond, dump_filter, strlen(dump_filter)))
+		return 0;
+
 	if (strncmp("pid/", cond, 4) == 0) {
 		svc_t *svc;
 
-		svc= client_svc_find_by_cond(cond);
+		svc = client_svc_find_by_cond(cond);
 		if (!svc) {
 			nm  = "unknown";
 			pid = 0;
@@ -308,7 +375,22 @@ static int dump_one_cond(const char *fpath, const struct stat *sb, int tflag, st
 		pid = 1;
 	}
 
-	printf("%-*d  %-*s  %-6s  <%s>\n", pw, pid, iw, nm, asserted, cond);
+	if (json) {
+		if (!dump_once)
+			puts("[");
+
+		printf("%s  {\n"
+		       "    \"pid\": %d,\n"
+		       "    \"identity\": \"%s\",\n"
+		       "    \"status\": \"%s\",\n"
+		       "    \"condition\": \"%s\"\n"
+		       "  }",
+		       dump_once ? ",\n" : "",
+		       pid, nm, asserted, cond);
+
+		dump_once++;
+	} else
+		printf("%-*d  %-*s  %-6s  <%s>\n", pw, pid, iw, nm, asserted, cond);
 
 	return 0;
 }
@@ -316,55 +398,116 @@ static int dump_one_cond(const char *fpath, const struct stat *sb, int tflag, st
 static int do_cond_dump(char *arg)
 {
 	col_widths();
-	if (heading) {
-		char title[80];
-
-		snprintf(title, sizeof(title), "%-*s  %-*s  %-6s  %s", pw, "PID",
-			 iw, "IDENT", "STATUS", "CONDITION");
-		print_header(title);
-	}
+	if (heading && !json)
+		print_header("%-*s  %-*s  %-6s  %s", pw, "PID", iw, "IDENT",
+			     "STATUS", "CONDITION");
 
+	dump_once = 0;
+	dump_filter = arg;
 	if (nftw(_PATH_COND, dump_one_cond, 20, 0) == -1) {
-		warnx("Failed parsing %s", _PATH_COND);
+		WARNX("Failed parsing %s", _PATH_COND);
 		return 1;
 	}
+	if (dump_once)
+		puts("\n]");
 
 	return 0;
 }
 
-static int do_cond_act(char *arg, int creat)
+typedef enum { COND_CLR, COND_SET, COND_GET } condop_t;
+
+static cond_state_t cond_read(char *path)
 {
-	char oneshot[256];
-	size_t off;
+	int now, gen;
 
-	if (arg && strncmp(arg, COND_USR, strlen(COND_USR)) == 0)
-		arg += strlen(COND_USR);
+	if (fngetint(path, &gen) == -1)
+		return COND_OFF;
 
-	if (!arg || !arg[0])
-		errx(1, "Invalid condition (empty)");
-	if (strchr(arg, '/'))
-		errx(1, "Invalid condition (slashes)");
-	if (strchr(arg, '.'))
-		errx(1, "Invalid condition (periods)");
+	/*
+	 * if we cannot read the reconf generation, then either sth is
+	 * very wrong, or we are called very early/late boot/shutdown.
+	 */
+	if (fngetint(_PATH_RECONF, &now) == -1)
+		return COND_FLUX;	/* classify as flux */
 
-	snprintf(oneshot, sizeof(oneshot), _PATH_CONDUSR "%s", arg);
-	off = strlen(_PATH_COND);
+	if (now != gen)
+		return COND_FLUX;
 
-	if (creat) {
-		if (symlink(_PATH_RECONF, oneshot) && errno != EEXIST)
-			err(1, "Failed asserting condition <%s>", &oneshot[off]);
-	} else {
-		if (erase(oneshot) && errno != ENOENT)
-			err(1, "Failed deasserting condition <%s>", &oneshot[off]);
+	return COND_ON;
+}
+
+/*
+ * cond get allows only one argument
+ * cond set|clr iterate over multiple args
+ */
+static int do_cond_act(char *args, condop_t op)
+{
+	cond_state_t cstate;
+	char path[256];
+	char *arg;
+
+	if (!args || !args[0])
+		ERRX(2, "Invalid condition (empty)");
+
+	arg = strtok(args, " \t");
+	while (arg) {
+		size_t off;
+
+		if (strncmp(arg, COND_USR, strlen(COND_USR)) == 0)
+			arg += strlen(COND_USR);
+
+		/* allowed to read any condition, but not set/clr */
+		if (op != COND_GET) {
+			if (strchr(arg, '/'))
+				ERRX(2, "Invalid condition (slashes)");
+			if (strchr(arg, '.'))
+				ERRX(2, "Invalid condition (periods)");
+		}
+
+		if (strchr(arg, '/'))
+			snprintf(path, sizeof(path), _PATH_COND "%s", arg);
+		else
+			snprintf(path, sizeof(path), _PATH_CONDUSR "%s", arg);
+		off = strlen(_PATH_COND);
+
+		switch (op) {
+		case COND_GET:
+			cstate = cond_read(path);
+			if (verbose)
+				puts(condstr(cstate));
+
+			switch (cstate) {
+			case COND_ON:
+				return 0;
+			case COND_OFF:
+				return 1;
+			case COND_FLUX:
+			default:
+				break;
+			}
+			return 255;
+
+		case COND_SET:
+			if (symlink(_PATH_RECONF, path) && errno != EEXIST)
+				ERR(73, "Failed asserting condition <%s>", &path[off]);
+			break;
+		case COND_CLR:
+			if (erase(path) && errno != ENOENT)
+				ERR(73, "Failed deasserting condition <%s>", &path[off]);
+			break;
+		}
+
+		arg = strtok(NULL, " \t");
 	}
 
 	return 0;
 }
 
-static int do_cond_set(char *arg) { return do_cond_act(arg, 1); }
-static int do_cond_clr(char *arg) { return do_cond_act(arg, 0); }
+static int do_cond_get(char *arg) { return do_cond_act(arg, COND_GET); }
+static int do_cond_set(char *arg) { return do_cond_act(arg, COND_SET); }
+static int do_cond_clr(char *arg) { return do_cond_act(arg, COND_CLR); }
 
-static char *svc_cond(svc_t *svc, char *buf, size_t len)
+static char *svc_cond(svc_t *svc, char *buf, size_t len, int ansi)
 {
 	char *cond, *conds;
 
@@ -377,11 +520,18 @@ static char *svc_cond(svc_t *svc, char *buf, size_t len)
 	if (!conds)
 		return buf;
 
-	strlcat(buf, "<", len);
+	if (json)
+		strlcat(buf, "[ ", len);
 
 	for (cond = strtok(conds, ","); cond; cond = strtok(NULL, ",")) {
-		if (cond != conds)
+		if (cond != conds) {
 			strlcat(buf, ",", len);
+			if (json)
+				strlcat(buf, " ", len);
+		}
+
+		if (json)
+			strlcat(buf, "\"", len);
 
 		switch (cond_get(cond)) {
 		case COND_ON:
@@ -390,26 +540,30 @@ static char *svc_cond(svc_t *svc, char *buf, size_t len)
 			break;
 
 		case COND_FLUX:
-			if (!plain)
+			if (ansi)
 				strlcat(buf, "\e[1m", len);
 			strlcat(buf, "~", len);
 			strlcat(buf, cond, len);
-			if (!plain)
+			if (ansi)
 				strlcat(buf, "\e[0m", len);
 			break;
 
 		case COND_OFF:
-			if (!plain)
+			if (ansi)
 				strlcat(buf, "\e[1m", len);
 			strlcat(buf, "-", len);
 			strlcat(buf, cond, len);
-			if (!plain)
+			if (ansi)
 				strlcat(buf, "\e[0m", len);
 			break;
 		}
+
+		if (json)
+			strlcat(buf, "\"", len);
 	}
 
-	strlcat(buf, ">", len);
+	if (json)
+		strlcat(buf, " ]", len);
 
 	return buf;
 }
@@ -418,24 +572,41 @@ static int do_cond_show(char *arg)
 {
 	enum cond_state cond;
 	char buf[512];
+	int once = 0;
 	svc_t *svc;
 
 	col_widths();
-	if (heading) {
-		char title[80];
-
-		snprintf(title, sizeof(title), "%-*s  %-*s  %-6s  %s", pw, "PID",
-			 iw, "IDENT", "STATUS", "CONDITION (+ ON, ~ FLUX, - OFF)");
-		print_header(title);
-	}
+	if (heading && !json)
+		print_header("%-*s  %-*s  %-6s  %s", pw, "PID", iw, "IDENT",
+			     "STATUS", "CONDITION (+ ON, ~ FLUX, - OFF)");
 
 	for (svc = client_svc_iterator(1); svc; svc = client_svc_iterator(0)) {
 		if (!svc->cond[0])
 			continue;
 
 		cond = cond_get_agg(svc->cond);
-
 		svc_ident(svc, buf, sizeof(buf));
+
+		if (json) {
+			if (!once)
+				puts("[");
+
+			printf("%s  {\n"
+			       "    \"pid\": %d,\n"
+			       "    \"identity\": \"%s\",\n"
+			       "    \"status\": \"%s\",\n",
+			       once ? ",\n" : "",
+			       svc->pid,
+			       buf,
+			       condstr(cond));
+			svc_cond(svc, buf, sizeof(buf), !plain);
+			printf("    \"condition\": %s\n"
+			       "  }", buf);
+
+			once++;
+			continue;
+		}
+
 		printf("%-*d  %-*s  ", pw, svc->pid, iw, buf);
 
 		if (cond == COND_ON)
@@ -443,17 +614,24 @@ static int do_cond_show(char *arg)
 		else
 			printf("\e[1m%-6.6s\e[0m  ", condstr(cond));
 
-		puts(svc_cond(svc, buf, sizeof(buf)));
+		svc_cond(svc, buf, sizeof(buf), !plain);
+		if (buf[0])
+			printf("<%s>", buf);
+		puts("");
 	}
 
+	if (json && once)
+		puts("\n]");
+
 	return 0;
 }
 
 static int do_cmd(int cmd)
 {
 	struct init_request rq = {
-		.magic = INIT_MAGIC,
-		.cmd = cmd,
+		.magic     = INIT_MAGIC,
+		.cmd       = cmd,
+		.sleeptime = timeout,
 	};
 
 	if (client_send(&rq, sizeof(rq))) {
@@ -540,6 +718,45 @@ static int show_version(char *arg)
 	return 0;
 }
 
+static int plugins_list(char *arg)
+{
+	struct init_request rq = {
+		.magic = INIT_MAGIC,
+		.cmd   = INIT_CMD_GET_PLUGINS,
+	};
+	char buf[sizeof(rq.data) + 1];
+	char *ptr;
+	int rc;
+
+	rc = client_send(&rq, sizeof(rq));
+	if (rc) {
+		if (rc == 255)
+			return 69;
+		else
+			ERRX(69, "No such command");
+	}
+	memcpy(buf, rq.data, sizeof(buf) - 1);
+	buf[sizeof(rq.data)] = 0;
+
+	if (heading)
+		print_header("%-18s  %s", "PLUGIN", "DEPENDENCIES");
+
+	ptr = strtok(buf, " ");
+	while (ptr) {
+		rq.cmd = INIT_CMD_PLUGIN_DEPS;
+		strlcpy(rq.data, ptr, sizeof(rq.data));
+		rc = client_send(&rq, sizeof(rq));
+		if (rc)
+			printf("%s\n", ptr);
+		else
+			printf("%-18s  %s\n", ptr, rq.data);
+
+		ptr = strtok(NULL, " ");
+	}
+
+	return 0;
+}
+
 /**
  * runlevel_string - Convert a bit encoded runlevel to .conf syntax
  * @levels: Bit encoded runlevels
@@ -550,18 +767,19 @@ static int show_version(char *arg)
  */
 char *runlevel_string(int currlevel, int levels)
 {
-	int i, pos = 1;
 	static char lvl[21];
+	int i = INIT_LEVEL;
+	int pos = 1;
 
 	memset(lvl, 0, sizeof(lvl));
 	lvl[0] = '[';
 
-	for (i = 0; i < 10; i++) {
+	do {
 		if (ISSET(levels, i)) {
 			if (!plain && currlevel == i)
 				pos = strlcat(lvl, "\e[1m", sizeof(lvl));
 
-			if (i == 0)
+			if (i == INIT_LEVEL)
 				lvl[pos++] = 'S';
 			else
 				lvl[pos++] = '0' + i;
@@ -571,7 +789,14 @@ char *runlevel_string(int currlevel, int levels)
 		} else {
 			lvl[pos++] = '-';
 		}
-	}
+
+		/* XXX: ugly hack to get order right: S0123456789 */
+		if (i == INIT_LEVEL)
+			i = 0;
+		else
+			i++;
+	} while (i < INIT_LEVEL);
+
 
 	lvl[pos++] = ']';
 	lvl[pos]   = 0;
@@ -579,6 +804,32 @@ char *runlevel_string(int currlevel, int levels)
 	return lvl;
 }
 
+char *runlevel_arr(int levels)
+{
+	static char lvl[42];
+	int i = INIT_LEVEL;
+	int p = 2, s = 0;
+
+	strlcpy(lvl, "[ ", sizeof(lvl));
+	do {
+		if (ISSET(levels, i)) {
+			if (i == INIT_LEVEL)
+				p += snprintf(&lvl[p], sizeof(lvl) - p, "\"S\"");
+			else
+				p += snprintf(&lvl[p], sizeof(lvl) - p, "%s%c", s ? ", " : "", '0' + i);
+			s++;
+		}
+		/* XXX: ugly hack to get order right: S0123456789 */
+		if (i == INIT_LEVEL)
+			i = 0;
+		else
+			i++;
+	} while (i < INIT_LEVEL);
+	strlcat(lvl, " ]", sizeof(lvl));
+
+	return lvl;
+}
+
 static int missing(svc_t *svc)
 {
 	if (svc->state == SVC_HALTED_STATE && svc_is_missing(svc))
@@ -587,11 +838,11 @@ static int missing(svc_t *svc)
 	return 0;
 }
 
-static char *svc_command(svc_t *svc, char *buf, size_t len)
+static char *svc_command(svc_t *svc, char *buf, size_t len, int ansi)
 {
-	int bold = missing(svc);
+	int bold = missing(svc) && ansi;
 
-	if (plain || whichp(svc->cmd))
+	if (whichp(svc->cmd))
 		bold = 0;
 
 	strlcpy(buf, bold ? "\e[1m" : "", len);
@@ -605,16 +856,23 @@ static char *svc_command(svc_t *svc, char *buf, size_t len)
 		strlcat(buf, svc->args[i], len);
 	}
 
+	if (svc_is_sysv(svc)) {
+		char *cmd = svc->state == SVC_HALTED_STATE ? "stop" : "start";
+
+		strlcat(buf, " ", len);
+		strlcat(buf, cmd, len);
+	}
+
 	strlcat(buf, bold ? "\e[0m" : "", len);
 
 	return buf;
 }
 
-static char *svc_environ(svc_t *svc, char *buf, size_t len)
+static char *svc_environ(svc_t *svc, char *buf, size_t len, int ansi)
 {
 	int bold = missing(svc);
 
-	if (plain || svc_checkenv(svc) || svc->env[0] == '-')
+	if (!ansi || svc_checkenv(svc) || svc->env[0] == '-')
 		bold = 0;
 
 	strlcpy(buf, bold ? "\e[1m" : "", len);
@@ -624,17 +882,24 @@ static char *svc_environ(svc_t *svc, char *buf, size_t len)
 	return buf;
 }
 
-static char *exit_status(int status, char *buf, size_t len)
+static char *exit_status(svc_t *svc, char *buf, size_t len)
 {
 	int rc, sig;
+	char *str;
 
-	rc = WEXITSTATUS(status);
-	sig = WTERMSIG(status);
+	rc = WEXITSTATUS(svc->status);
+	sig = WTERMSIG(svc->status);
 
-	if (WIFEXITED(status))
-		snprintf(buf, len, " (code=exited, status=%d%s)", rc, code2str(rc));
-	else if (WIFSIGNALED(status))
-		snprintf(buf, len, " (code=signal, status=%d%s)", sig, sig2str(sig));
+	if (WIFEXITED(svc->status)) {
+		str = code2str(rc);
+		snprintf(buf, len, " (code=exited, status=%d%s%s%s)", rc,
+			 str[0] ? "/" : "", str,
+			 svc->manual ? ", manual=yes" : "");
+	}
+	else if (WIFSIGNALED(svc->status)) {
+		str = sig2str(sig);
+		snprintf(buf, len, " (code=signal, status=%d%s%s)", sig, str[0] ? "/" : "", str);
+	}
 
 	return buf;
 }
@@ -649,7 +914,7 @@ static char *status(svc_t *svc, int full)
 	s = svc_status(svc);
 	switch (svc->state) {
 	case SVC_HALTED_STATE:
-		exit_status(svc->status, ok, sizeof(ok));
+		exit_status(svc, ok, sizeof(ok));
 		color = "\e[1m";
 		break;
 
@@ -658,7 +923,7 @@ static char *status(svc_t *svc, int full)
 		break;
 
 	case SVC_DONE_STATE:
-		exit_status(svc->status, ok, sizeof(ok));
+		exit_status(svc, ok, sizeof(ok));
 		if (WIFEXITED(svc->status)) {
 			if (WEXITSTATUS(svc->status))
 				color = "\e[1;31m";
@@ -673,7 +938,7 @@ static char *status(svc_t *svc, int full)
 		break;
 
 	default:
-		exit_status(svc->status, ok, sizeof(ok));
+		exit_status(svc, ok, sizeof(ok));
 		color = "\e[1;33m";
 		break;
 	}
@@ -705,6 +970,106 @@ static void show_cgroup_tree(char *group, char *pfx)
 	cgroup_tree(path, pfx, 0, 0);
 }
 
+/*
+ * arg: 'foo'   should match foo:1 foo:2, etc. but not foobar
+ * arg: 'foo:1' should only match foo:1
+ * arg: 'foo:'  is allowed to fail, unsupported syntax atm
+ * arg: 'foo:*' is allowed to fail, unsupported syntax atm
+ */
+static int svc_compare(svc_t *svc, char *arg)
+{
+	char ident[MAX_IDENT_LEN];
+	char *ptr;
+
+	svc_ident(svc, ident, sizeof(ident));
+	ptr = strchr(ident, ':');
+	if (ptr && !strchr(arg, ':'))
+		*ptr = 0;
+
+	if (!strcmp(ident, arg))
+		return 1;
+
+	return 0;
+}
+
+static int json_status_one(FILE *fp, svc_t *svc, char *indent, int prev)
+{
+	long now = jiffies();
+	char *pidfn = NULL;
+	char buf[512];
+
+	pidfn = svc->pidfile;
+	if (pidfn[0] == '!')
+		pidfn++;
+	else if (pidfn[0] == 0)
+		pidfn = "none";
+
+	fprintf(fp,
+		"%s"
+		"%s{\n"
+		"%s  \"identity\": \"%s\",\n"
+		"%s  \"description\": \"%s\",\n"
+		"%s  \"type\": \"%s\",\n"
+		"%s  \"forking\": %s,\n"
+		"%s  \"status\": \"%s\",\n",
+		prev ? ",\n" : indent, prev ? indent : "",
+		indent, svc_ident(svc, NULL, 0),
+		indent, svc->desc,
+		indent, svc_typestr(svc),
+		indent, svc->forking ? "true" : "false",
+		indent, svc_status(svc));
+	if (svc->state != SVC_RUNNING_STATE) {
+		int rc, sig;
+
+		rc = WEXITSTATUS(svc->status);
+		sig = WTERMSIG(svc->status);
+
+		if (WIFEXITED(svc->status))
+			fprintf(fp,
+				"%s  \"exit\": { \"%s\": %d },\n",
+				indent, "code", rc);
+		else if (WIFSIGNALED(svc->status))
+			fprintf(fp,
+				"%s  \"exit\": { \"%s\": %d },\n",
+				indent, "signal", sig);
+	}
+	fprintf(fp,
+		"%s  \"origin\": \"%s\",\n"
+		"%s  \"command\": \"%s\",\n",
+		indent, svc->file[0] ? svc->file : "built-in",
+		indent, svc_command(svc, buf, sizeof(buf), 0));
+	svc_environ(svc, buf, sizeof(buf), 0);
+	if (buf[0])
+		fprintf(fp,
+			"%s  \"environment\": \"%s\",\n", indent, buf);
+	svc_cond(svc, buf, sizeof(buf), 0);
+	if (buf[0])
+		fprintf(fp,
+			"%s  \"condition\": %s,\n", indent, buf);
+	if (svc->manual)
+		fprintf(fp,
+			"%s  \"starts\": %d,\n", indent, svc->once);
+	fprintf(fp,
+		"%s  \"restarts\": %d,\n", indent, svc->restart_tot); /* XXX: add restart_cnt and restart_max */
+	fprintf(fp,
+		"%s  \"pidfile\": \"%s\",\n"
+		"%s  \"pid\": %d,\n"
+		"%s  \"user\": \"%s\",\n"
+		"%s  \"group\": \"%s\",\n"
+		"%s  \"uptime\": %ld,\n"
+		"%s  \"runlevels\": %s\n"
+		"%s}",
+		indent, pidfn,
+		indent, svc->pid,
+		indent, svc->username,
+		indent, svc->group,
+		indent, svc->pid ? now - svc->start_time : 0,
+		indent, runlevel_arr(svc->runlevels),
+		indent);
+
+	return 0;
+}
+
 static int show_status(char *arg)
 {
 	char ident[MAX_IDENT_LEN];
@@ -718,25 +1083,33 @@ static int show_status(char *arg)
 		long now = jiffies();
 		char uptm[42] = "N/A";
 		char *pidfn = NULL;
-		int exact = 0;
 
-		for (svc = client_svc_iterator(1); svc; svc = client_svc_iterator(0)) {
-			svc_ident(svc, ident, sizeof(ident));
-			if (string_match(ident, arg))
-				num++;
-			if (string_case_compare(ident, arg))
-				exact++;
-		}
+		for (svc = client_svc_iterator(1); svc; svc = client_svc_iterator(0))
+			num += svc_compare(svc, arg);
 
-		if (num > 1 && !exact)
+		if (num > 1)
 			break;
 
 		svc = client_svc_find(arg);
 		if (!svc)
-			return 255;
+			ERRX(noerr ? 0 : 69, "no such task or service(s): %s", arg);
 
-		if (quiet)
+		if (quiet) {
+			if (svc_is_runtask(svc)) {
+				if (svc->started)
+					return 0;
+				return 1;
+			}
 			return svc->state != SVC_RUNNING_STATE;
+		}
+
+		if (json) {
+			int rc;
+
+			rc = json_status_one(stdout, svc, "", 0);
+			puts("");
+			return rc;
+		}
 
 		pidfn = svc->pidfile;
 		if (pidfn[0] == '!')
@@ -748,14 +1121,20 @@ static int show_status(char *arg)
 		printf("   Identity : %s\n", svc_ident(svc, ident, sizeof(ident)));
 		printf("Description : %s\n", svc->desc);
 		printf("     Origin : %s\n", svc->file[0] ? svc->file : "built-in");
-		printf("Environment : %s\n", svc_environ(svc, buf, sizeof(buf)));
-		printf("Condition(s): %s\n", svc_cond(svc, buf, sizeof(buf)));
-		printf("    Command : %s\n", svc_command(svc, buf, sizeof(buf)));
+		svc_environ(svc, buf, sizeof(buf), !plain);
+		if (buf[0])
+			printf("Environment : %s\n", buf);
+		svc_cond(svc, buf, sizeof(buf), !plain);
+		if (buf[0])
+			printf("Condition(s): <%s>\n", buf);
+		printf("    Command : %s\n", svc_command(svc, buf, sizeof(buf), !plain));
 		printf("   PID file : %s\n", pidfn);
 		printf("        PID : %d\n", svc->pid);
 		printf("       User : %s\n", svc->username);
 		printf("      Group : %s\n", svc->group);
 		printf("     Uptime : %s\n", svc->pid ? uptime(now - svc->start_time, uptm, sizeof(uptm)) : uptm);
+		if (svc->manual)
+			printf("     Starts : %d\n", svc->once);
 		printf("   Restarts : %d (%d/%d)\n", svc->restart_tot, svc->restart_cnt, svc->restart_max);
 		printf("  Runlevels : %s\n", runlevel_string(runlevel, svc->runlevels));
 		if (cgrp && svc->pid > 1) {
@@ -765,6 +1144,8 @@ static int show_status(char *arg)
 			char *group;
 
 			group = pid_cgroup(svc->pid, grbuf, sizeof(grbuf));
+			if (!group)
+				goto no_cgroup; /* ... or PID doesn't exist (anymore) */
 			snprintf(path, sizeof(path), "%s/%s", FINIT_CGPATH, group);
 			cg = cg_conf(path);
 
@@ -774,30 +1155,49 @@ static int show_status(char *arg)
 			       cg->cg_mem.min, cg->cg_mem.max);
 			show_cgroup_tree(group, "              ");
 		}
+	no_cgroup:
 		printf("\n");
 
-		return do_log(svc->cmd);
+		return do_log(svc, "| tail -10");
+	}
+
+	if (json) {
+		int prev = 0;
+
+		for (svc = client_svc_iterator(1); svc; svc = client_svc_iterator(0)) {
+			svc_ident(svc, ident, sizeof(ident));
+			if (num && !svc_compare(svc, arg))
+				continue;
+
+			if (!prev)
+				fputs("[\n", stdout);
+			json_status_one(stdout, svc, "  ", prev++);
+		}
+		if (prev)
+			fputs("\n]\n", stdout);
+
+		return 0;
 	}
 
 	col_widths();
 	if (heading) {
 		char title[80];
 
-		snprintf(title, sizeof(title), "%-*s  %-*s  %-8s %-12s ",
+		snprintf(title, sizeof(title), "%-*s  %-*s  %-8s %-13s ",
 			 pw, "PID", iw, "IDENT", "STATUS", "RUNLEVELS");
 		if (!verbose)
 			strlcat(title, "DESCRIPTION", sizeof(title)); 
 		else
 			strlcat(title, "COMMAND", sizeof(title)); 
 
-		print_header(title);
+		print_header("%s", title);
 	}
 
 	for (svc = client_svc_iterator(1); svc; svc = client_svc_iterator(0)) {
 		char *lvls;
 
 		svc_ident(svc, ident, sizeof(ident));
-		if (num && !string_match(ident, arg))
+		if (num && !svc_compare(svc, arg))
 			continue;
 
 		printf("%-*d  ", pw, svc->pid);
@@ -805,14 +1205,14 @@ static int show_status(char *arg)
 
 		lvls = runlevel_string(runlevel, svc->runlevels);
 		if (strchr(lvls, '\e'))
-			printf("%-20.20s ", lvls);
+			printf("%-21.21s ", lvls);
 		else
-			printf("%-12.12s ", lvls);
+			printf("%-13.13s ", lvls);
 
 		if (!verbose)
 			puts(svc->desc);
 		else
-			puts(svc_command(svc, buf, sizeof(buf)));
+			puts(svc_command(svc, buf, sizeof(buf), !plain));
 	}
 
 	return 0;
@@ -860,9 +1260,9 @@ static int transform(char *nm)
 
 static int has_conf(char *path, size_t len, char *name)
 {
-	paste(path, len, FINIT_RCSD, name);
+	paste(path, len, finit_rcsd, name);
 	if (!fisdir(path)) {
-		strlcpy(path, FINIT_RCSD, len);
+		strlcpy(path, finit_rcsd, len);
 		return 0;
 	}
 
@@ -871,7 +1271,7 @@ static int has_conf(char *path, size_t len, char *name)
 
 static int usage(int rc)
 {
-	int has_rcsd = fisdir(FINIT_RCSD);
+	int has_rcsd = fisdir(finit_rcsd);
 	int has_ena;
 	char avail[256];
 	char ena[256];
@@ -886,31 +1286,34 @@ static int usage(int rc)
 		"  -b, --batch               Batch mode, no screen size probing\n"
 		"  -c, --create              Create missing paths (and files) as needed\n"
 		"  -f, --force               Ignore missing files and arguments, never prompt\n"
+		"  -h, --help                This help text\n"
+		"  -j, --json                JSON output in 'status' and 'cond' commands\n"
+		"  -n, --noerr               Ignore error, e.g., already started/enabled/...\n"
 		"  -1, --once                Only one lap in commands like 'top'\n"
 		"  -p, --plain               Use plain table headings, no ctrl chars\n"
 		"  -q, --quiet               Silent, only return status of command\n"
 		"  -t, --no-heading          Skip table headings\n"
 		"  -v, --verbose             Verbose output\n"
-		"  -h, --help                This help text\n"
+		"  -V, --version             Show program version\n"
 		"\n"
 		"Commands:\n"
 		"  debug                     Toggle Finit (daemon) debug\n"
 		"  help                      This help text\n"
-		"  version                   Show Finit version\n"
+		"  version                   Show program version\n"
 		"\n", prognm);
 
 	if (has_rcsd)
 		fprintf(stderr,
-			"  ls | list                 List all .conf in " FINIT_RCSD "\n"
+			"  ls | list                 List all .conf in %s\n"
 			"  create   <CONF>           Create   .conf in %s\n"
 			"  delete   <CONF>           Delete   .conf in %s\n"
 			"  show     <CONF>           Show     .conf in %s\n"
 			"  edit     <CONF>           Edit     .conf in %s\n"
 			"  touch    <CONF>           Change   .conf in %s\n",
-			avail, avail, avail, avail, avail);
+			finit_rcsd, avail, avail, avail, avail, avail);
 	else
 		fprintf(stderr,
-			"  show                      Show     %s\n", FINIT_CONF);
+			"  show                      Show     %s\n", finit_conf);
 
 	if (has_ena)
 		fprintf(stderr,
@@ -920,24 +1323,27 @@ static int usage(int rc)
 			"  disable  <CONF>           Disable  .conf in %s\n", ena);
 	if (has_rcsd)
 		fprintf(stderr,
-			"  reload                    Reload  *.conf in " FINIT_RCSD " (activate changes)\n");
+			"  reload                    Reload  *.conf in %s (activate changes)\n", finit_rcsd);
 	else
 		fprintf(stderr,
-			"  reload                    Reload " FINIT_CONF " (activate changes)\n");
+			"  reload                    Reload   %s (activate changes)\n", finit_conf);
 
 	fprintf(stderr,
-//		"  reload   <NAME>[:ID]      Reload (SIGHUP) service by name\n"  
 		"\n"
-		"  cond     set   <COND>     Set (assert) user-defined condition     +usr/COND\n"
-		"  cond     clear <COND>     Clear (deassert) user-defined condition -usr/COND\n"
+		"  cond     set   <COND>     Set (assert) user-defined conditions     +usr/COND\n"
+		"  cond     get   <COND>     Get status of user-defined condition, see $? and -v\n"
+		"  cond     clear <COND>     Clear (deassert) user-defined conditions -usr/COND\n"
 		"  cond     status           Show condition status, default cond command\n"
-		"  cond     dump             Dump all conditions and their status\n"
+		"  cond     dump  [TYPE]     Dump all, or a type of, conditions and their status\n"
 		"\n"
 		"  log      [NAME]           Show ten last Finit, or NAME, messages from syslog\n"
 		"  start    <NAME>[:ID]      Start service by name, with optional ID\n"
 		"  stop     <NAME>[:ID]      Stop/Pause a running service by name\n"
-		"  reload   <NAME>[:ID]      Reload service by name (SIGHUP or restart)\n"
+		"  reload   <NAME>[:ID]      Reload service as if .conf changed (SIGHUP or restart)\n"
+		"                            This allows restart of run/tasks that have already run\n"
+		"                            Note: Finit .conf file(s) are *not* reloaded!\n"
 		"  restart  <NAME>[:ID]      Restart (stop/start) service by name\n"
+		"  signal   <NAME>[:ID] <S>  Send signal S to service by name, with optional ID\n"
 		"  ident    [NAME]           Show matching identities for NAME, or all\n"
 		"  status   <NAME>[:ID]      Show service status, by name\n"
 		"  status                    Show status of services, default command\n");
@@ -948,6 +1354,8 @@ static int usage(int rc)
 			"  ps                        List processes based on cgroups\n"
 			"  top                       Show top-like listing based on cgroups\n");
 	fprintf(stderr,
+		"\n"
+		"  plugins                   List installed plugins\n"
 		"\n"
 		"  runlevel [0-9]            Show or set runlevel: 0 halt, 6 reboot\n"
 		"  reboot                    Reboot system\n"
@@ -994,12 +1402,15 @@ static int cmd_parse(int argc, char *argv[], struct cmd *command)
 		if (!cmd_cond(&command[i]))
 			continue;
 
-		if (!string_match(command[i].cmd, argv[0]))
+		if (!string_compare(command[i].cmd, argv[0]))
 			continue;
 
 		if (command[i].ctx)
 			return cmd_parse(argc - 1, &argv[1], command[i].ctx);
 
+		if (command[i].cb_multiarg)
+			return command[i].cb_multiarg(argc - 1, argv + 1);
+
 		if (command[i].cb) {
 			int rc = 0;
 
@@ -1014,10 +1425,54 @@ static int cmd_parse(int argc, char *argv[], struct cmd *command)
 	}
 
 	if (argv[0] && strlen(argv[0]) > 0)
-		errx(1, "No such command.  See 'initctl help' for an overview of available commands.");
+		ERRX(3, "No such command.  See 'initctl help' for an overview of available commands.");
 
 	return command[0].cb(NULL); /* default cmd */
 }
+
+static void cleanup(void)
+{
+	if (finit_conf)
+		free(finit_conf);
+	if (finit_rcsd)
+		free(finit_rcsd);
+}
+
+static void sourcerc(void)
+{
+	char *line;
+	FILE *fp;
+
+	fp = fopen(_PATH_VARRUN "finit/.initrc", "r");
+	if (!fp)
+		goto fallback;
+
+	while ((line = fparseln(fp, NULL, NULL, NULL, 0))) {
+		char *val;
+
+		if ((val = fgetval(line, "FINIT_CONF", "="))) {
+			if (finit_conf)
+				free(finit_conf);
+			finit_conf = val;
+		}
+		if ((val = fgetval(line, "FINIT_RCSD", "="))) {
+			if (finit_rcsd)
+				free(finit_rcsd);
+			finit_rcsd = val;
+		}
+
+		free(line);
+	}
+
+fallback:
+	if (!finit_conf)
+		finit_conf = strdup(FINIT_CONF);
+	if (!finit_rcsd)
+		finit_rcsd = strdup(FINIT_RCSD);
+
+	atexit(cleanup);
+}
+
 int main(int argc, char *argv[])
 {
 	struct option long_options[] = {
@@ -1026,60 +1481,69 @@ int main(int argc, char *argv[])
 		{ "debug",      0, NULL, 'd' },
 		{ "force",      0, NULL, 'f' },
 		{ "help",       0, NULL, 'h' },
+		{ "json",       0, NULL, 'j' },
+		{ "noerr",      0, NULL, 'n' },
 		{ "once",       0, NULL, '1' },
 		{ "plain",      0, NULL, 'p' },
 		{ "quiet",      0, NULL, 'q' },
 		{ "no-heading", 0, NULL, 't' },
 		{ "verbose",    0, NULL, 'v' },
+		{ "version",    0, NULL, 'V' },
 		{ NULL, 0, NULL, 0 }
 	};
 	struct cmd cond[] = {
-		{ "status",   NULL, do_cond_show, NULL }, /* default cmd */
-		{ "dump",     NULL, do_cond_dump, NULL },
-		{ "set",      NULL, do_cond_set,  NULL },
-		{ "clr",      NULL, do_cond_clr,  NULL },
-		{ "clear",    NULL, do_cond_clr,  NULL },
-		{ NULL, NULL, NULL, NULL }
+		{ "status",   NULL, do_cond_show, NULL, NULL }, /* default cmd */
+		{ "dump",     NULL, do_cond_dump, NULL, NULL  },
+		{ "set",      NULL, do_cond_set,  NULL, NULL  },
+		{ "get",      NULL, do_cond_get,  NULL, NULL  },
+		{ "clr",      NULL, do_cond_clr,  NULL, NULL  },
+		{ "clear",    NULL, do_cond_clr,  NULL, NULL  },
+		{ NULL, NULL, NULL, NULL, NULL  }
 	};
 	struct cmd command[] = {
-		{ "status",   NULL, show_status,  NULL }, /* default cmd */
-		{ "ident",    NULL, show_ident,   NULL },
+		{ "status",   NULL, show_status,  NULL, NULL  }, /* default cmd */
+		{ "ident",    NULL, show_ident,   NULL, NULL  },
 
-		{ "debug",    NULL, toggle_debug, NULL },
-		{ "devel",    NULL, do_devel,     NULL },
-		{ "help",     NULL, do_help,      NULL },
-		{ "version",  NULL, show_version, NULL },
+		{ "debug",    NULL, toggle_debug, NULL, NULL  },
+		{ "devel",    NULL, do_devel,     NULL, NULL  },
+		{ "help",     NULL, do_help,      NULL, NULL  },
+		{ "version",  NULL, show_version, NULL, NULL  },
 
-		{ "list",     NULL, serv_list,    NULL },
-		{ "ls",       NULL, serv_list,    NULL },
-		{ "enable",   NULL, serv_enable,  NULL },
-		{ "disable",  NULL, serv_disable, NULL },
-		{ "touch",    NULL, serv_touch,   NULL },
-		{ "show",     NULL, serv_show,    NULL },
-		{ "edit",     NULL, serv_edit,    NULL },
-		{ "create",   NULL, serv_creat,   NULL },
-		{ "delete",   NULL, serv_delete,  NULL },
-		{ "reload",   NULL, do_reload,    NULL },
+		{ "list",     NULL, serv_list,    NULL, NULL  },
+		{ "ls",       NULL, serv_list,    NULL, NULL  },
+		{ "enable",   NULL, serv_enable,  NULL, NULL  },
+		{ "disable",  NULL, serv_disable, NULL, NULL  },
+		{ "touch",    NULL, serv_touch,   NULL, NULL  },
+		{ "show",     NULL, serv_show,    NULL, NULL  },
+		{ "cat",      NULL, serv_show,    NULL, NULL  }, /* alias */
+		{ "edit",     NULL, serv_edit,    NULL, NULL  },
+		{ "create",   NULL, serv_creat,   NULL, NULL  },
+		{ "delete",   NULL, serv_delete,  NULL, NULL  },
+		{ "reload",   NULL, do_reload,    NULL, NULL  },
 
-		{ "cond",     cond, NULL, NULL         },
+		{ "cond",     cond, NULL, NULL, NULL          },
 
-		{ "log",      NULL, do_log,       NULL },
-		{ "start",    NULL, do_start,     NULL },
-		{ "stop",     NULL, do_stop,      NULL },
-		{ "restart",  NULL, do_restart,   NULL },
+		{ "log",      NULL, show_log,     NULL, NULL  },
+		{ "start",    NULL, do_start,     NULL, NULL  },
+		{ "stop",     NULL, do_stop,      NULL, NULL  },
+		{ "restart",  NULL, do_restart,   NULL, NULL  },
+		{ "signal",   NULL, NULL,         NULL, do_signal  },
+		{ "kill",     NULL, NULL,         NULL, do_signal  }, /* alias */
 
-		{ "cgroup",   NULL, show_cgroup, &cgrp },
-		{ "ps",       NULL, show_cgps,   &cgrp },
-		{ "top",      NULL, show_cgtop,  &cgrp },
+		{ "cgroup",   NULL, show_cgroup, &cgrp, NULL  },
+		{ "ps",       NULL, show_cgps,   &cgrp, NULL  },
+		{ "top",      NULL, show_cgtop,  &cgrp, NULL  },
 
-		{ "runlevel", NULL, do_runlevel,  NULL },
-		{ "reboot",   NULL, do_reboot,    NULL },
-		{ "halt",     NULL, do_halt,      NULL },
-		{ "poweroff", NULL, do_poweroff,  NULL },
-		{ "suspend",  NULL, do_suspend,   NULL },
+		{ "plugins",  NULL, plugins_list, NULL, NULL  },
 
-		{ "utmp",     NULL, do_utmp,     &utmp },
-		{ NULL, NULL, NULL, NULL }
+		{ "runlevel", NULL, do_runlevel,  NULL, NULL  },
+		{ "reboot",   NULL, do_reboot,    NULL, NULL  },
+		{ "halt",     NULL, do_halt,      NULL, NULL  },
+		{ "poweroff", NULL, do_poweroff,  NULL, NULL  },
+		{ "suspend",  NULL, do_suspend,   NULL, NULL  },
+
+		{ "utmp",     NULL, do_utmp,     &utmp, NULL  },
+		{ NULL, NULL, NULL, NULL, NULL  }
 	};
 	int interactive = 1, c;
 
@@ -1087,10 +1551,11 @@ int main(int argc, char *argv[])
 		return reboot_main(argc, argv);
 
 	/* Enable functionality depending on system capabilities */
+	sourcerc();
 	cgrp = cgroup_avail();
 	utmp = has_utmp();
 
-	while ((c = getopt_long(argc, argv, "1bcdfh?pqtv", long_options, NULL)) != EOF) {
+	while ((c = getopt_long(argc, argv, "1bcdfh?jnpqtvV", long_options, NULL)) != EOF) {
 		switch(c) {
 		case '1':
 			ionce = 1;
@@ -1116,6 +1581,14 @@ int main(int argc, char *argv[])
 		case '?':
 			return usage(0);
 
+		case 'j':
+			json = 1;
+			break;
+
+		case 'n':
+			noerr = 1;
+			break;
+
 		case 'p':
 			plain = 1;
 			break;
@@ -1131,6 +1604,9 @@ int main(int argc, char *argv[])
 		case 'v':
 			verbose = 1;
 			break;
+
+		case 'V':
+			return show_version(NULL);
 		}
 	}
 
@@ -1144,6 +1620,9 @@ void logit(int prio, const char *fmt, ...)
 {
 	va_list ap;
 
+	if (quiet)
+		return;
+
 	va_start(ap, fmt);
 	if (prio <= LOG_ERR)
 		verrx(1, fmt, ap);
