@@ -1,6 +1,6 @@
 /* reboot, halt, poweroff, shutdown, suspend for finit
  *
- * Copyright (c) 2009-2021  Joachim Wiberg <troglobit@gmail.com>
+ * Copyright (c) 2009-2024  Joachim Wiberg <troglobit@gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -26,6 +26,7 @@
 #include <getopt.h>
 #include <signal.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include "util.h"
@@ -42,6 +43,8 @@ static cmd_t cmd = CMD_UNKNOWN;
 static char *msg = NULL;
 
 /* initctl API */
+extern int timeout;
+
 extern int do_reboot  (char *arg);
 extern int do_halt    (char *arg);
 extern int do_poweroff(char *arg);
@@ -98,21 +101,17 @@ static void do_kill(int signo, char *msg)
 	do_sleep(5);
 }
 
-static void sig(int signo)
-{
-	(void)signo;		/* NOP */
-}
-
 static int usage(int rc)
 {
 	fprintf(stderr, "Usage: %s [OPTIONS]\n\n"
 		"Options:\n"
-		"      --help      This help text\n"
-		"  -f, --force     Force unsafe %s now, do not contact the init system.\n"
-		"      --halt      Halt system, regardless of how the command is called.\n"
-		"  -h              Halt or power off after shutdown.\n"
-		"  -P, --poweroff  Power-off system, regardless of how the command is called.\n"
-		"  -r, --reboot    Reboot system, regardless of how the command is called.\n"
+		"  -?, --help         This help text\n"
+		"  -f, --force        Force unsafe %s now, do not contact the init system.\n"
+		"      --halt         Halt system, regardless of how the command is called.\n"
+		"  -h                 Halt or power off after shutdown.\n"
+		"  -P, --poweroff     Power-off system, regardless of how the command is called.\n"
+		"  -r, --reboot       Reboot system, regardless of how the command is called.\n"
+		"  -t, --timeout=SEC  Force reboot/shutdown after a given timeout\n"
 		"\n", prognm, msg);
 
 	return rc;
@@ -120,20 +119,22 @@ static int usage(int rc)
 
 int reboot_main(int argc, char *argv[])
 {
-	int c, force = 0;
 	struct option long_options[] = {
-		{"help",     0, NULL, 'h'},
+		{"help",     0, NULL, '?'},
 		{"force",    0, NULL, 'f'},
 		{"halt",     0, NULL, 'H'},
 		{"poweroff", 0, NULL, 'p'},
 		{"reboot",   0, NULL, 'r'},
+		{"timeout",  1, NULL, 't'},
 		{NULL, 0, NULL, 0}
 	};
+	int c, force = 0;
+	int tmp;
 
 	/* Initial command taken from program name */
 	transform(prognm);
 
-	while ((c = getopt_long(argc, argv, "h?fHPpr", long_options, NULL)) != EOF) {
+	while ((c = getopt_long(argc, argv, "h?fHPprt:", long_options, NULL)) != EOF) {
 		switch(c) {
 		case '?':
 			return usage(0);
@@ -156,6 +157,13 @@ int reboot_main(int argc, char *argv[])
 			cmd = CMD_REBOOT;
 			break;
 
+		case 't':
+			tmp = atoi(optarg);
+			if (tmp < 1)
+				errx(1, "Invalid forced timeout value.");
+			timeout = tmp;
+			break;
+
 		default:
 			return usage(1);
 		}
@@ -164,10 +172,13 @@ int reboot_main(int argc, char *argv[])
 	/* Check for any overrides */
 	transform(NULL);
 
-	/* Catch and ignore SIGTERM on halt/poweroff */
-	signal(SIGTERM, sig);
+	/* Check if we're in sulogin shell */
+	if (getenv("SULOGIN"))
+		force = 1;
 
 	if (force) {
+		sync();
+
 		switch (cmd) {
 		case CMD_REBOOT:
 			c = reboot(RB_AUTOBOOT);
@@ -182,9 +193,9 @@ int reboot_main(int argc, char *argv[])
 			break;
 
 		case CMD_SUSPEND:
-			c = reboot(RB_SW_SUSPEND);
+			c = suspend();
 			if (c && errno == EINVAL)
-				errx(1, "Kernel does not support suspend.");
+				errx(1, "Kernel does not support suspend to RAM.");
 			break;
 
 		case CMD_UNKNOWN:
