@@ -1,7 +1,7 @@
 /* Finit TTY handling
  *
  * Copyright (c) 2013       Mattias Walström <lazzer@gmail.com>
- * Copyright (c) 2013-2021  Joachim Wiberg <troglobit@gmail.com>
+ * Copyright (c) 2013-2024  Joachim Wiberg <troglobit@gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -48,7 +48,7 @@ char	*tty_canonicalize (char *dev);
 int	 tty_isatcon      (char *dev);
 char	*tty_atcon	  (void);
 
-int	 tty_parse_args   (char *cmdline, struct tty *tty);
+int	 tty_parse_args   (struct tty *tty, char *cmd, char **args);
 
 int	 tty_exists	  (char *dev);
 int	 tty_exec	  (svc_t *tty);
